!= ͥǡեץ
!
!= dcpam main program for generation of initial data file
!
! Authors::   Yasuhiro MORIKAWA
!             modified by Shin-ichi Takehiro for venus experiments of primitive model.
! Version::   $Id: init_data_primitive_venus.f90,v 1.4 2010-09-10 10:03:22 takepiro Exp $
! Tag Name::  $Name:  $
! Copyright:: Copyright (C) GFD Dennou Club, 2008. All rights reserved.
! License::   See COPYRIGHT[link:../../../COPYRIGHT]
!

program init_data
  !
  ! <b>Note that Japanese and English are described in parallel.</b>
  !
  ! ͥǡեޤ. 
  !
  ! Initial data file is created. 
  !

  ! ⥸塼 ; USE statements
  !

  ! ʻ
  ! Grid points settings
  !
  use gridset, only: imax, & ! ٳʻ. 
                             ! Number of grid points in longitude
    &                jmax, & ! ٳʻ. 
                             ! Number of grid points in latitude
    &                kmax    ! ľؿ. 
                             ! Number of vertical level
  ! ɸǡ
  ! Axes data settings
  !
  use axesset, only: &
    & x_Lon, &
                            ! $ \lambda $ [rad.] . . Longitude
    & y_Lat, &
                            ! $ \varphi $ [rad.] . . Latitude
    & z_Sigma
                            ! $ \sigma $ ٥ (). 
                            ! Full $ \sigma $ level

  ! 
  ! Time control
  !
  use timeset, only: &
    & TimeN                   ! ƥå $ t $ λ. 
                              ! Time of step $ t $. 

  ! ꥹȥǡ
  ! Restart data input/output
  !
  use restart_file_io, only: RestartFileOutPut

  ! ҥȥǡ
  ! History data output
  !
  use gtool_historyauto, only: HistoryAutoPut

  ! ٤Ⱦҥ٥, ȹ٤λ
  ! Interpolate temperature on half sigma level, 
  ! and calculate pressure and height
  !
  use auxiliary, only: AuxVars

  ! ̷ѥ᥿
  ! Kind type parameter
  !
  use dc_types, only: DP, &      ! ټ¿. Double precision. 
    &                 STRING, &  ! ʸ.       Strings. 
    &                 TOKEN      ! .   Keywords. 

  ! ñ׻Τζ
  ! forcing for simple Venus calculation
  !
  use venus_simple_forcing, only: VenusSimpleNCTempEq

  ! ʸ ; Declaration statements
  !
  implicit none

  ! ͽѿ (ƥå $ t-\Delta t $ , $ t $ , $ t+\Delta t $ )
  ! Prediction variables  (Step $ t-\Delta t $ , $ t $ , $ t+\Delta t $ )
  !
  real(DP), allocatable:: xyz_U (:,:,:)
                              ! $ u $ .   ®. Eastward wind
  real(DP), allocatable:: xyz_V (:,:,:)
                              ! $ v $ .   ®. Northward wind
  real(DP), allocatable:: xyz_Temp (:,:,:)
                              ! $ T $ .   . Temperature
  real(DP), allocatable:: xy_Ps (:,:)
                              ! $ p_s $ . ɽ̵. Surface pressure

  real(DP), allocatable:: xyz_TempB (:,:,:)
                              ! $ T (t-\Delta t) $ .   . Temperature (K)
  real(DP), allocatable:: xyz_TempN (:,:,:)
                              ! $ T (t) $ .     . Temperature (K)

  ! ѿ, ¾
  ! Diagnostic variables, etc.
  !
  real(DP), allocatable:: xy_SurfHeight (:,:)
                              ! $ z_s $ . ɽ̹ (m)
                              ! Surface height (m)
  real(DP), allocatable:: xyz_Height (:,:,:)
                              !  (٥) (m)
                              ! Height (full level) (m)
  ! ѿ
  ! Work variables
  !
  integer:: i                 ! ˲ DO 롼Ѻѿ
                              ! Work variables for DO loop in longitude
  integer:: j                 ! ˲ DO 롼Ѻѿ
                              ! Work variables for DO loop in latitude
  integer:: k                 ! ľ˲ DO 롼Ѻѿ
                              ! Work variables for DO loop in vertical direction
  integer:: n                 ! ˲ DO 롼Ѻѿ
                              ! Work variables for DO loop in dimension of constituents

  real(DP):: eps              ! ٤θ
  real(DP),parameter:: &
             eps_null=1.0d-10
                              ! ٤θǤ뤳ȤȽ

  real(DP):: UInit            ! Ϳ®ʬ
			      ! The initial uniform longitudinal velocity
  real(DP):: TempInit         ! ͿͲ
			      ! The initial uniform temp. 
  real(DP):: PsInit           ! Ϳɽ̵
			      ! The initial uniform surface pressure. 
  real(DP):: TempDist_Amp     ! ϥǡͿ벹پο
			      ! The amplitude of temp. dist. of the output data


  ! NAMELIST ѿ
  ! NAMELIST group name
  !
  namelist /init_data_venus_nml/ UInit, TempInit, PsInit, TempDist_Amp


  ! ¹ʸ ; Executable statement
  !

  ! ץν (֥롼)
  ! Initialization for the main program (Internal subroutine)
  !
  call MainInit

  ! ·׻
  ! relaxation calculation
  !
  do while ( eps < eps_null )

    ! ٤Ⱦҥ٥, ȹ٤λ
    ! Interpolate temperature on half sigma level,
    ! and calculate pressure and height
    !
    call AuxVars( &
      & xy_Ps,      xyz_TempB,                               & ! (in )
      & xy_SurfHeight = xy_SurfHeight,                       & ! (in ) optional
      & xyz_Height    = xyz_Height                           & ! (out) optional
      & )

    call VenusSimpleNCTempEq( &
         & xyz_Height, &
         & xyz_TempN   &
         & )

    eps = maxval(abs(xyz_TempN - xyz_TempB))

    ! λ֥ƥåפ˸ͽѿؤ
    ! Exchange prediction variables for the next time step
    !
    xyz_TempB  = xyz_TempN

  ! ·׻λ
  ! Relaxation calculation is finished
  !
  end do
  !
  ! ҥȥǡ
  ! History data output
  !
  call HistoryAutoPut( TimeN, 'Temp',   xyz_TempB )
  call HistoryAutoPut( TimeN, 'Height', xyz_Height )

  xyz_Temp = xyz_TempB

  ! ٤˾Ϳ
  ! Add perturbation to temperature
  !
  do k = 1, kmax
    do j = 1, jmax
      do i = 0, imax - 1
        xyz_Temp(i,j,k) = &
          &    xyz_Temp(i,j,k) &
          &  + TempDist_Amp * sin ( x_Lon(i) * y_Lat(j) ) &
          &  - TempDist_Amp * ( 1.0_DP - z_Sigma(k) )
      end do
    end do
  end do

  ! ͥǡν
  ! Output initial data
  !
  call RestartFileOutput( &
    & xyz_U, xyz_V, xyz_Temp, xy_Ps )  ! (in)

  ! ץνλ (֥롼)
  ! Termination for the main program (Internal subroutine)
  !
  call MainTerminate

contains

  !-------------------------------------------------------------------

  subroutine MainInit
    !
    ! ץν³. 
    !
    ! Initialization procedure for the main program. 
    !

    ! MPI
    !
    use mpi_wrapper, only : MPIWrapperInit

    !
    use dc_message, only: MessageNotify

    use dc_types, only: &
      & STDOUT               ! ɸϤֹ. Unit number of standard output
    ! ե
    ! File I/O support
    !
    use dc_iounit, only: FileOpen

    ! ޥɥ饤
    ! Command line option parser
    !
    use option_parser, only: OptParseInit

    ! NAMELIST եϤ˴ؤ桼ƥƥ
    ! Utilities for NAMELIST file input
    !
    use namelist_util, only: NmlutilInit, NmlutilMsg, namelist_filename

    ! 
    ! Time control
    !
    use timeset, only: TimesetInit
                              ! ƥå $ t $ λ. Time of step $ t $. 

    ! ϥեδܾ
    ! Management basic information for output files
    ! 
    use fileset, only: FilesetInit

    ! ʻ
    ! Grid points settings
    !
    use gridset, only: GridsetInit, &
      &                imax, & ! ٳʻ. 
                               ! Number of grid points in longitude
      &                jmax, & ! ٳʻ. 
                               ! Number of grid points in latitude
      &                kmax    ! ľؿ. 
                               ! Number of vertical level

    ! ʪ
    ! Physical constants settings
    !
    use constants, only: ConstantsInit

    ! ɸǡ
    ! Axes data settings
    !
    use axesset, only: AxessetInit

    ! ꥹȥǡ
    ! Restart data input/output
    !
    use restart_file_io, only: RestartFileOpen, RestartFileGet

    ! ҥȥǡ
    ! History data output
    !
    use history_file_io, only: HistoryFileOpen
    use gtool_historyauto, only: HistoryAutoAddVariable

    ! 
    ! Time control
    !
    use timeset, only: TimesetInit, TimesetDelTimeHalf, &
      & TimeN                 ! ƥå $ t $ λ. Time of step $ t $. 

    ! ʸ ; Declaration statements
    !
    implicit none
    character(*), parameter:: prog_name = 'init_data_primitive_venus'
                              ! ץ̾. 
                              ! Main program name
    character(*), parameter:: version = &
      & '$Name:  $' // &
      & '$Id: init_data_primitive_venus.f90,v 1.4 2010-09-10 10:03:22 takepiro Exp $'
                              ! ץΥС
                              ! Main program version
    character(STRING):: brief
                              ! ¹ԥեδʷ
                              ! Brief account of executable file

    integer:: unit_nml        ! NAMELIST ե륪ץֹ. 
                              ! Unit number for NAMELIST file open
    integer:: iostat_nml      ! NAMELIST ɤ߹߻ IOSTAT. 
                              ! IOSTAT of NAMELIST read

    ! ¹ʸ ; Executable statement
    !

    ! Initialize MPI
    !
    call MPIWrapperInit

    brief = 'Initial data generation'
    call MessageNotify( 'M', prog_name, 'Run: %c', c1 = trim(brief) )
    call MessageNotify( 'M', prog_name, '-- version = %c', c1 = trim(version) )

    ! ޥɥ饤
    ! Command line option parser
    !
    call OptParseInit(prog_name, brief)

    ! NAMELIST ե̾
    ! Input NAMELIST file name
    !
    call NmlutilInit

    UInit        = 400.0d0
    TempInit     = 400.0d0
    PsInit       = 9.2d6
    TempDist_Amp = 1.0d-2

    ! NAMELIST init_data_venus_nml
    !
    if ( trim(namelist_filename) /= '' ) then
      call FileOpen( unit_nml, &          ! (out)
        & namelist_filename, mode = 'r' ) ! (in)

      TempDist_Amp = 0.0_DP
      rewind( unit_nml )
      read( unit_nml, &            ! (in)
           & nml = init_data_venus_nml, &  ! (out)
           & iostat = iostat_nml )    ! (out)
      close( unit_nml )

      call NmlutilMsg( iostat_nml, prog_name ) ! (in)
      if ( iostat_nml == 0 ) write( STDOUT, nml = init_data_venus_nml )
    end if

    ! 
    ! Time control
    !
    call TimesetInit

    ! ϥեδܾ
    ! Management basic information for output files
    ! 
    call FilesetInit

    ! ʻ
    ! Grid points settings
    !
    call GridsetInit

    ! ʪ
    ! Physical constants settings
    !
    call ConstantsInit

    ! ɸǡ
    ! Axes data settings
    !
    call AxessetInit

    ! ͽѿγ
    ! Allocation of prediction variables
    !
    allocate( xyz_U    (0:imax-1, 1:jmax, 1:kmax) )
    allocate( xyz_V    (0:imax-1, 1:jmax, 1:kmax) )
    allocate( xyz_Temp (0:imax-1, 1:jmax, 1:kmax) )
    allocate( xy_Ps    (0:imax-1, 1:jmax) )


    ! ͽѿγ
    ! Allocation of prediction variables
    !
    allocate( xyz_TempB (0:imax-1, 1:jmax, 1:kmax) )
    allocate( xyz_TempN (0:imax-1, 1:jmax, 1:kmax) )

    ! ѿγ
    ! Allocation of diagnostic variables
    !
    allocate( xy_SurfHeight (0:imax-1, 1:jmax) )
    allocate( xyz_Height (0:imax-1, 1:jmax, 1:kmax) )

    ! ͥǡ
    ! Initial data output
    !
    call RestartFileOpen( flag_init_data = .true. ) ! (in) optional

    ! ҥȥǡեν
    ! Initialization of history data files
    !
    call HistoryFileOpen

    ! ҥȥǡϤΤΤؤѿϿ
    ! Register of variables for history data output
    !
    call HistoryAutoAddVariable( 'Temp' , &      ! (in)
      & (/ 'lon ', 'lat ', 'sig ', 'time' /), &  ! (in)
      & 'temperature', 'K' )                     ! (in)
    call HistoryAutoAddVariable( 'Height', &
      & (/ 'lon ', 'lat ', 'sig ', 'time' /), &
      & 'Height', 'm' )

    ! ·׻
    !
    xy_SurfHeight = 0.0D0
    xyz_TempB = TempInit

    xyz_U     = UInit
    xyz_V     = 0.0d0
    xy_Ps     = PsInit

  end subroutine MainInit

  !-------------------------------------------------------------------

  subroutine MainTerminate
    !
    ! ץνλ³. 
    !
    ! Termination procedure for the main program. 
    !

    ! MPI
    !
    use mpi_wrapper, only : MPIWrapperFinalize

    ! 
    ! Time control
    !
    use timeset, only: TimesetClose

    ! ꥹȥǡ
    ! Restart data input/output
    !
    use restart_file_io, only: RestartFileClose

    ! ҥȥǡ
    ! History data output
    !
    use history_file_io, only: HistoryFileClose

    ! ʸ ; Declaration statements
    !
    implicit none

    ! ¹ʸ ; Executable statement
    !

    ! ꥹȥǡե륯
    ! Close restart data input
    !
    call RestartFileClose

    ! λ
    ! Termination of time control
    !
    call TimesetClose

    ! ͽѿγղ
    ! Deallocation of prediction variables
    !
    deallocate( xyz_U    )
    deallocate( xyz_V    )
    deallocate( xyz_Temp )
    deallocate( xy_Ps    )

    ! ҥȥǡե륯
    ! Close history data files
    !
    call HistoryFileClose

    ! Finalize MPI
    !
    call MPIWrapperFinalize

  end subroutine MainTerminate

end program init_data
